# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Sheepdog's clientside code.

This code is typically only run on the worker, and this file is currently
only used by pasting it into a job file (as workers don't generally have
sheepdog itself installed).
"""


import time
import json
import requests
import traceback

try:
    deserialise_function
except NameError:
    from sheepdog.serialisation import (deserialise_function,
                                        deserialise_arg,
                                        deserialise_namespace,
                                        serialise_pickle)

class Client:
    """Find out what to do, do it, report back."""
    HTTP_RETRIES = 10

    def __init__(self, url, request_id, job_index):
        self.url = url
        self.request_id = request_id
        self.job_index = job_index

    def get_details(self):
        """Retrieve the function to run and arguments to run with from the
           server.
        """
        url = self.url + "?request_id={0}&job_index={1}"
        url = url.format(self.request_id, self.job_index)
        tries = 0
        while tries < self.HTTP_RETRIES:
            try:
                r = requests.get(url)
                break
            except (requests.ConnectionError, requests.Timeout):
                tries += 1
                time.sleep(1)
                continue
        if tries == self.HTTP_RETRIES:
            raise RuntimeError("Could not connect to server.")
       
        r = r.json()
        self.args = deserialise_arg(r['args'])
        self.ns = deserialise_namespace(r['ns'])
        self.func = deserialise_function(r['func'], self.ns)

    def run(self):
        """Run the downloaded function, storing the result."""
        if not hasattr(self, 'func') or not hasattr(self, 'args'):
            raise RuntimeError("Must call `get_details` before `run`.")
        try:
            self.result = self.func(*self.args)
        except:
            self._submit_error(traceback.format_exc())

    def submit_results(self):
        if not hasattr(self, 'result'):
            raise RuntimeError("Must call `run` before `submit_results`.")
        result = serialise_pickle(self.result)
        self._submit(self.url, dict(result=result))

    def _submit_error(self, error):
        self._submit(self.url + "error", dict(error=str(error)))

    def _submit(self, url, data):
        data.update(
            {"request_id": self.request_id, "job_index": self.job_index})
        tries = 0
        while tries < self.HTTP_RETRIES:
            try:
                r = requests.post(url, data=data)
                break
            except (requests.ConnectionError, requests.Timeout):
                tries += 1
                time.sleep(1)
                continue
        if tries == self.HTTP_RETRIES:
            raise RuntimeError("Could not submit to server.")

    def go(self):
        """Call get_details(), run(), submit_results().
           Just for convenience.
        """
        self.get_details()
        self.run()
        if hasattr(self, 'result'):
            self.submit_results()



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Interface to the storage backend.

Future plans involve porting most of those handwritten SQL to a sensible ORM.
"""

import sqlite3

schema = """
CREATE TABLE IF NOT EXISTS requests (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    function BLOB,
    namespace BLOB,
    date_submitted TEXT
);

CREATE TABLE IF NOT EXISTS tasks (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    request_id INTEGER,
    job_index INTEGER,
    args BLOB,
    FOREIGN KEY (request_id) REFERENCES requests(id)
);

CREATE TABLE IF NOT EXISTS results (
    id INTEGER PRIMARY KEY,
    task_id INTEGER,
    result BLOB,
    FOREIGN KEY (task_id) REFERENCES tasks(id)
);

CREATE TABLE IF NOT EXISTS errors (
    id INTEGER PRIMARY KEY,
    task_id INTEGER,
    error TEXT,
    FOREIGN KEY (task_id) REFERENCES tasks(id)
);
"""

class Storage:
    """Manage persistence for requests and results.

    Request functions and result objects are stored as binary
    blobs in the database, so any bytes object will be fine.
    They'll be returned as they were sent.

    """

    def __init__(self, dbfile="./sheepdog.sqlite"):
        """__init__ creates a database connection.

        dbfile is a file path for the sqlite file, or `:memory:` to only
        use in memory persistence.
        """
        self.dbfile = dbfile
        self.conn = sqlite3.connect(dbfile)

    def initdb(self):
        """Create the database structure if it doesn't already exist.
        """
        c = self.conn.cursor()
        c.executescript(schema)
        self.conn.commit()

    def new_request(self, serialised_function, serialised_namespace,
                    args_list):
        """Add a new request to the database.

        serialised_function is some bytes object that should be given to
        workers to turn into the code to execute.

        serialised_namespace is some bytes object that should be given to
        workers alongside the serialised function to provide helper variables
        and functions that the primary function requires.

        args_list is a list, tuple or other iterable where each item is
        some bytes object that should be given to workers to run their
        target function with.

        Returns the new request ID.
        """
        c = self.conn.cursor()
        c.execute("INSERT INTO requests (function, namespace, date_submitted)"
                  " VALUES (?, ?, date('now'))",
                  (sqlite3.Binary(serialised_function),
                   sqlite3.Binary(serialised_namespace)))
        request_id = c.lastrowid
        tasks_list = []
        for idx, arg in enumerate(args_list):
            tasks_list.append((request_id, idx + 1, sqlite3.Binary(arg)))
        c.executemany("INSERT INTO tasks (request_id, job_index, args)"
                      " VALUES (?, ?, ?)", tasks_list)
        self.conn.commit()
        return request_id

    def get_details(self, request_id, job_index):
        """Get the target function, namespace and arguments for a given job.
        """
        c = self.conn.cursor()
        c.execute("SELECT requests.function, requests.namespace, tasks.args"
                  " FROM tasks"
                  " JOIN requests ON tasks.request_id=requests.id"
                  " WHERE request_id=? AND job_index=?",
                  (request_id, job_index))
        details = c.fetchone()
        if not details:
            raise ValueError("No details found for specified request and job.")
        return (bytes(details[0]), bytes(details[1]), bytes(details[2]))

    def _get_task_id(self, request_id, job_index):
        """Retrieve the task ID for a given request ID and job index."""
        c = self.conn.cursor()
        c.execute("SELECT id FROM tasks"
                  " WHERE request_id=? AND job_index=?",
                  (request_id, job_index))
        task_id = c.fetchone()
        if not task_id:
            raise ValueError("No task found for specified request and job.")
        return task_id[0]

    def store_result(self, request_id, job_index, result):
        """Store a new result from a given request_id and job_index."""
        task_id = self._get_task_id(request_id, job_index)
        c = self.conn.cursor()
        c.execute("INSERT INTO results (task_id, result) VALUES (?, ?)",
                  (task_id, sqlite3.Binary(result)))
        self.conn.commit()

    def store_error(self, request_id, job_index, error):
        """Store an error resulting from a computation."""
        task_id = self._get_task_id(request_id, job_index)
        c = self.conn.cursor()
        c.execute("INSERT INTO errors (task_id, error) VALUES (?, ?)",
                  (task_id, error))
        self.conn.commit()

    def count_results(self, request_id):
        """Count the number of results so far for the given request_id."""
        c = self.conn.cursor()
        c.execute("SELECT COUNT(*)"
                  " FROM results"
                  " JOIN tasks ON results.task_id=tasks.id"
                  " WHERE tasks.request_id=?", (request_id,))
        return c.fetchone()[0]

    def count_errors(self, request_id):
        """Count the number of errors reported so far."""
        c = self.conn.cursor()
        c.execute("SELECT COUNT(*)"
                  " FROM errors"
                  " JOIN tasks ON errors.task_id=tasks.id"
                  " WHERE tasks.request_id=?", (request_id,))
        return c.fetchone()[0]

    def count_results_and_errors(self, request_id):
        """Sum the result and error counts."""
        return self.count_results(request_id) + self.count_errors(request_id)

    def count_tasks(self, request_id):
        """Count the total number of tasks for this request."""
        c = self.conn.cursor()
        c.execute("SELECT COUNT(*) FROM tasks WHERE tasks.request_id=?",
                  (request_id,))
        return c.fetchone()[0]

    def get_results(self, request_id):
        """Fetch all results for a given request_id.

        Returns a list of (args, result) items in the order of the original
        args_list provided to new_request.

        Gaps are not filled in, so if results have not yet been submitted
        the corresponding arguments will not appear in this list and this
        list will be shorter than the length of args_list.
        """
        c = self.conn.cursor()
        c.execute("SELECT tasks.args, results.result"
                  " FROM results"
                  " JOIN tasks ON results.task_id=tasks.id"
                  " WHERE tasks.request_id=?"
                  " ORDER BY tasks.job_index", (request_id,))
        return [(bytes(r[0]), bytes(r[1])) for r in c.fetchall()]

    def get_errors(self, request_id):
        """Fetch all errors for a given request_id.

        Returns a list of (args, error) items in the order of the original
        args_list provided to new_request.
        """
        c = self.conn.cursor()
        c.execute("SELECT tasks.args, errors.error"
                  " FROM errors"
                  " JOIN tasks ON errors.task_id=tasks.id"
                  " WHERE tasks.request_id=?"
                  " ORDER BY tasks.job_index", (request_id,))
        return [(bytes(r[0]), r[1]) for r in c.fetchall()]



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Sheepdog is a utility to run arbitary code on a GridEngine cluster and collect
the results, typically by mapping a set of arguments to one function.

Documentation: http://sheepdog.readthedocs.org

Source code: https://github.com/adamgreig/sheepdog

PyPI: https://pypi.python.org/pypi/Sheepdog

Sheepdog is released under the MIT license, see the LICENSE file for details.
"""

__author__ = "Adam Greig <adam@adamgreig.com>"
__version__ = "0.1.4"
__version_info__ = tuple([int(d) for d in __version__.split(".")])
__license__ = "MIT License"

import os
import sys
import time
import copy
import socket
import getpass

from sheepdog.server import Server
from sheepdog.storage import Storage
from sheepdog.deployment import deploy_and_run
from sheepdog.job_file import job_file

from sheepdog import serialisation

default_config = {
    "ssh_port": 22,
    "ssh_user": getpass.getuser(),
    "ssh_dir": "$HOME/.sheepdog",
    "dbfile": "./sheepdog.sqlite",
    "port": 7676,
    "ge_opts": ["-r y", "-l ubuntu=1", "-l lr=0", "-wd $HOME/.sheepdog/",
                "-o $HOME/.sheepdog/", "-e $HOME/.sheepdog/"],
    "shell": "/usr/bin/python",
    "localhost": socket.getfqdn()
}


def map_sync(f, args, config, ns=None):
    """Run *f* with each of *args* on GridEngine and return the results.

       Optionally *ns* is a dict containing a namespace to execute the function
       in, which may itself contain additional functions.
       
       Blocks until all results are in.

       *config* must be a dict including:

            `host`: the hostname to submit grid engine jobs to [required]

            `ssh_port`: the ssh port to connect on
                        (default: 22)

            `ssh_user`: the ssh username to use
                        (default: current username)

            `ssh_dir`: the remote directory to put job scripts in
                       (default $HOME/.sheepdog)

            `dbfile`: the filename for the results db
                      (default ./sheepdog.sqlite)

            `port`: the port for the server to listen on
                    (default: 7676)

            `ge_opts`: a list of grid engine options
                       (default: ["-r y", "-l ubuntu=1", "-l lr=0",
                                  "-wd $HOME/.sheepdog/",
                                  "-o $HOME/.sheepdog/",
                                  "-e $HOME/.sheepdog/"])

            `shell`: the path to the python to run the job with
                     (default: "/usr/bin/python")

            `localhost`: the hostname for workers to find the local host
                         (default: system's FQDN)
    """
    if not ns:
        ns = {}

    conf = copy.copy(default_config)
    conf.update(config)
    func_bin = serialisation.serialise_function(f)
    args_bin = serialisation.serialise_args(args)
    namespace_bin = serialisation.serialise_namespace(ns)
    storage = Storage(dbfile=conf['dbfile'])
    storage.initdb()
    request_id = storage.new_request(func_bin, namespace_bin, args_bin)
    server = Server(port=conf['port'], dbfile=conf['dbfile'])
    url = "http://{0}:{1}/".format(conf['localhost'], conf['port'])
    n_args = len(args)
    jf = job_file(url, request_id, n_args, conf['shell'], conf['ge_opts'])
    print("Deploying job with request ID {0}...".format(request_id))
    deploy_and_run(conf['host'], jf, request_id,
                   conf['ssh_user'], conf['ssh_port'], conf['ssh_dir'])

    n_results = 0
    while n_results != n_args:
        n_results = storage.count_results(request_id)
        sys.stdout.write("Received {0}/{1} results...\r".format(
            n_results, n_args))
        sys.stdout.flush()
        time.sleep(1)

    return [serialisation.deserialise_pickle(r[1])
            for r in storage.get_results(request_id)]



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Sheepdog's HTTP server endpoints.

The Server class sets up a server on another subprocess, ready to receive
requests from workers. Uses Tornado if available, else falls back to the Flask
debug web server.
"""

import json
from multiprocessing import Process
from flask import Flask, request, g
from sheepdog.storage import Storage

try:
    from tornado.wsgi import WSGIContainer
    from tornado.httpserver import HTTPServer
    from tornado.ioloop import IOLoop
    USE_TORNADO = True
except ImportError:
    USE_TORNADO = False

app = Flask(__name__)

@app.route('/', methods=['GET'])
def get_config():
    """Endpoint for workers to fetch their configuration before execution.
       Workers should specify `request_id` (integer) and `job_index` (integer)
       from their job file.

       Returns a JSON object:
       
       {"func": (serialised function object),
        "args": (serialised arguments list)
       }

       with HTTP status 200 on success.
    """
    storage = get_storage()
    request_id = int(request.args['request_id'])
    job_index = int(request.args['job_index'])
    details = storage.get_details(request_id, job_index)
    func = details[0].decode()
    ns = details[1].decode()
    args = details[2].decode()
    return json.dumps({"func": func, "ns": ns, "args": args})

@app.route('/', methods=['POST'])
def submit_result():
    """Endpoint for workers to submit results arising from successful function
       execution. Should specify `request_id` (integer), `job_index` (integer)
       and `result` (serialised result) HTTP POST parameters.

       Returns the string "OK" and HTTP 200 on success.
    """
    storage = get_storage()
    request_id = int(request.form['request_id'])
    job_index = int(request.form['job_index'])
    result = request.form['result'].encode()
    storage.store_result(request_id, job_index, result)
    return "OK"

@app.route('/error', methods=['POST'])
def report_error():
    """Endpoint for workers to report back errors in function execution.
       Workers should specify `request_id` (integer), `job_index` (integer) and
       `error` (an error string) HTTP POST parameters.

       Returns the string "OK" and HTTP 200 on success.
    """
    storage = get_storage()
    request_id = int(request.form['request_id'])
    job_index = int(request.form['job_index'])
    error = str(request.form['error'])
    storage.store_error(request_id, job_index, error)
    return "OK"

def get_storage():
    """Retrieve the request-local database connection, creating it if required.
    """
    if not hasattr(g, '_storage'):
        dbfile = app.config['DBFILE']
        g._storage = Storage(dbfile) if dbfile else Storage()
    return g._storage

def run_server(port=7676, dbfile=None):
    """Start up the HTTP server. If Tornado is available it will be used, else
       fall back to the Flask debug server.
    """
    app.config['DBFILE'] = dbfile
    if USE_TORNADO:
        HTTPServer(WSGIContainer(app)).listen(port)
        IOLoop.instance().start()
    else:
        app.run(host='0.0.0.0', port=port)


class Server:
    """Run the HTTP server for workers to request arguments and return results.
    """

    def __init__(self, port=7676, dbfile=None):
        """__init__ creates and starts the HTTP server.
        """
        self.server = Process(target=run_server, args=(port, dbfile))
        self.server.start()

    def stop(self):
        """Terminate the HTTP server."""
        self.server.terminate()
        self.server.join()

    def __del__(self):
        self.stop()



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Code for deploying code to servers and executing jobs on GridEngine.

Right now this is a little unwieldly and uses SSH as a subprocess.
"""

# All I want for Christmas is Python 3 support in Paramiko.
# Until then...

import os
import tempfile
import subprocess

def deploy_and_run(host, jobfile, request_id, user, port, directory):
    """Deploy *jobfile* to *user*@*host*:*port* in *directory*,
       using *request_id* to name it, then submit it to the GridEngine cluster.
    """
    hoststr = "{0}@{1}".format(user, host)
    scpportstr = "-P {0}".format(port)
    sshportstr = "-p {0}".format(port)
    filestr = "{0}:{1}/sheepdog_{2}.py".format(hoststr, directory, request_id)
    mkdirstr = "mkdir -p {0}".format(directory)
    submitstr = "qsub {0}/sheepdog_{1}.py".format(directory, request_id)

    rv = subprocess.call(["ssh", sshportstr, hoststr, mkdirstr])

    if rv:
        raise RuntimeError("Error creating {0}".format(directory))

    fd, fp = tempfile.mkstemp()
    os.write(fd, jobfile.encode())

    rv = subprocess.call(["scp", scpportstr, fp, filestr])

    os.close(fd)
    os.unlink(fp)

    if rv:
        raise RuntimeError("Error uploading job file.")

    rv = subprocess.call(["ssh", sshportstr, hoststr, submitstr])

    if rv:
        raise RuntimeError("Error submitting job request.")



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Sheepdog's serialisation functions, for turning Python bits into
HTTP safe strings.

Functions are serialised by getting their code objects and dumping them with
marshal and encoding the bytes using base64.

Arguments are pickled since that should work for any standard arguments.
"""

import types
import base64
import pickle
import marshal

def serialise_function(f):
    """Turn a Python function (unbound, no closures, etc) into a base64 byte
       string (a bytes object).
    """
    if hasattr(f, "__code__"):
        fcode = f.__code__
    else:
        fcode = f.func_code
    fcodebin = marshal.dumps(fcode)
    return base64.b64encode(fcodebin)

def deserialise_function(f, namespace=None):
    """Turn a base64 bytestring back into a Python function, with *namespace*
       for its globals. Returns the function.
    """
    if not namespace:
        namespace = globals()
    else:
        namespace = namespace.copy()
    if "__builtins__" not in namespace:
        namespace["__builtins__"] = __builtins__
    fcodebin = base64.b64decode(f)
    fcode = marshal.loads(fcodebin)
    return types.FunctionType(fcode, namespace)

def serialise_pickle(args):
    """Serialise *args* using pickle and base64, returning the b64 bytestring.
    """
    return base64.b64encode(pickle.dumps(args))

def deserialise_pickle(args):
    """Deserialise *args* using base64 and pickle, returning the Python object.
    """
    return pickle.loads(base64.b64decode(args))

serialise_arg = serialise_pickle
deserialise_arg = deserialise_pickle

def serialise_args(args):
    """Serialise each item in *args* using serialise_pickle, returning a list
       of serialised items.
    """
    return [serialise_pickle(x) for x in args]

def deserialise_args(args):
    """Deserialise each item in *args* using deserialise_pickle, returning a
       list of the original objects.
    """
    return [deserialise_pickle(x) for x in args]

def serialise_namespace(ns):
    """Serialise a dict *ns* so that any functions in it are first converted to
       base64 strings, then pickle and base64 encode the result.
    """
    ns = ns.copy()
    serialised_keys = []
    for key in ns:
        if type(ns[key]) == types.FunctionType:
            ns[key] = serialise_function(ns[key])
            serialised_keys.append(key)
    ns["__serialised_keys"] = serialised_keys
    return serialise_pickle(ns)

def deserialise_namespace(ns):
    """Deserialise a base64 bytestring *ns*, deserialising any serialised
       functions inside it too, finally returning the original dictionary.
    """
    ns = deserialise_pickle(ns)
    if "__serialised_keys" in ns:
        for key in ns["__serialised_keys"]:
            ns[key] = deserialise_function(ns[key], ns)
    del ns["__serialised_keys"]
    return ns



from setuptools import setup

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ''

def get_version():
    with open("sheepdog/__init__.py") as f:
        for line in f:
            if line.startswith("__version__"):
                return line[15:-2]
    raise Exception("Could not find version number")

setup(
    name='Sheepdog',
    version=get_version(),
    author='Adam Greig',
    author_email='adam@adamgreig.com',
    packages=['sheepdog'],
    url='http://sheepdog.readthedocs.org',
    license='MIT',
    description='Shepherd GridEngine',
    long_description=long_description,
    test_suite='nose.collector',
    tests_require=['nose', 'requests'],
    install_requires=[
        "Flask >= 0.10.1",
        "requests >= 2.0.1",
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Topic :: Scientific/Engineering',
    ],
)



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Generate job files to send to the cluster.

The template is filled in with the job specifics and the formatted string is
returned ready for deployment.
"""

import inspect
from sheepdog import client, serialisation

template = """#!{shell}
{geopts}

# Sheepdog Job File
# Autogenerated by Sheepdog. Don't edit by hand.
# See https://github.com/adamgreig/sheepdog

###########################################################
## Serialisation code from Sheepdog:

{serialisation_code}

##
###########################################################

###########################################################
## Clientside code from Sheepdog:

{client_code}

##
###########################################################

import os
job_index = os.environ['SGE_TASK_ID']
Client("{url}", {request_id}, job_index).go()
"""

def job_file(url, request_id, n_args, shell, grid_engine_opts):
    """Format the template for a specific job, ready for deployment.
       
       *url* is the URL (including port) that the workers should contact to
       fetch job information, including a trailing slash.

       *request_id* is the request ID workers should use to associate
       themselves with the correct request.

       *n_args* is the number of jobs that will be queued in the array task,
       the same as the number of arguments being mapped by sheepdog.

       *shell* is the path to the Python that will execute the job. Could be a
       system or user Python, so long as it meets the Sheepdog requirements.
       Is used for the -S option to GridEngine as well as the script shebang.

       *grid_engine_opts* is a list of string arguments to Grid Engine to
       specify options such as resource requirements. 
    """
    grid_engine_opts.append("-t 1-{0}".format(n_args))
    grid_engine_opts.append("-S \"{0}\"".format(shell))
    geopts = '\n'.join("#$ {0}".format(opt) for opt in grid_engine_opts)
    client_code = inspect.getsource(client)
    serialisation_code = inspect.getsource(serialisation)
    return template.format(**locals())
